import tkinter

from tksugar.widgets.generatorsupport import GeneratorSupport
from tksugar.eventreciever import EventReciever

class Menu(tkinter.Menu, GeneratorSupport):
  """
  `tkinter.Menu` with added methods to support Generator.
  """
  def __init__(self, master=None, cnf={}, **kw):
    """Construct menu widget with the parent MASTER.

    Valid resource names: activebackground, activeborderwidth,
    activeforeground, background, bd, bg, borderwidth, cursor,
    disabledforeground, fg, font, foreground, postcommand, relief,
    selectcolor, takefocus, tearoff, tearoffcommand, title, type."""
    super().__init__(master=None, cnf=cnf, **kw)
    self._command = None
    self._parent = None
    if issubclass(type(master), tkinter.Wm):
      master.config(menu=self)

  def append_child(self, child, **params):
    params["menu"] = child
    child._parent = self
    self.add_cascade(params)

  def items(self, items):
    """
    Define a menu item.

    Parameters
    ----
    items: list(str or dict)
      An array of menu items. Information defining a string or menu item.
    """
    def radio(a):
      for n, i in enumerate(a["items"]):
        self.add_radiobutton(
          label= i["label"] if type(i) is dict else i,
          variable=a.get("variable", None),
          value= i.get("value", n) if type(i) is dict else n,
          command= a.get("command", None))
    def cascade(a):
      items = a.pop("items")
      tearoff = a.pop("tearoff", False)
      m = Menu(master=self, tearoff=tearoff)
      m.items(items)
      self.append_child(m, **a)
    for item in items:
      if type(item) is str:
        if item == "---":
          item = {
            "type": "separator"
          }
        else:
          item = {
            "type": "command",
            "label": item
          }
      item.setdefault("type", "command")
      if "name" in item and not "command" in item: item["command"] = EventReciever(self, item["name"], self._callback)
      if "label" in item and not "command" in item: item["command"] = EventReciever(self, item["label"], self._callback)
      if "name" in item: item.pop("name")
      switch = {
        "separator": lambda a: self.add_separator(),
        "command": lambda a: self.add_command(a),
        "check": lambda a: self.add_checkbutton(a),
        "radio": lambda a: radio(a),
        "cascade": lambda a: cascade(a)
      }
      t = item.pop("type")
      v = switch.get(t, ValueError)(item)
      if issubclass(type(v), Exception):
        raise v

  def _callback(self, o, n):
    """
    Callbacks for various menus.
    """
    if not self.command is None:
      self.command.tag.tag["item"] = n
      self.command()
    elif not self.parent is None:
      self.parent._callback(o, n)

  @property
  def parent(self):
    """
    Gets parent.
    """
    return self._parent

  @property
  def command(self):
    """
    Gets or sets an event handler that will be executed when the button is pressed.
    """
    return self._command

  @command.setter
  def command(self, value):
    """
    Gets or sets an event handler that will be executed when the button is pressed.
    """
    self._command = value



import unittest
import tkinter
import yaml

from tksugar.generator import Generator

class Button(tkinter.Button):
  """
  We override the Button class for testing.
  """
  def __init__(self, master=None, cnf={}, **kw):
    """Construct a button widget with the parent MASTER.

    STANDARD OPTIONS

        activebackground, activeforeground, anchor,
        background, bitmap, borderwidth, cursor,
        disabledforeground, font, foreground
        highlightbackground, highlightcolor,
        highlightthickness, image, justify,
        padx, pady, relief, repeatdelay,
        repeatinterval, takefocus, text,
        textvariable, underline, wraplength

    WIDGET-SPECIFIC OPTIONS

        command, compound, default, height,
        overrelief, state, width
    """
    tkinter.Button.__init__(self, **kw)
    self.items = []
    self.testdict = {}

  def items(self, items):
    self.items.append(items)

class Test_Generator(unittest.TestCase):
  """
  Tests the `Generator#generate()` method.
  """

  def tearDown(self):
    tkinter._default_root = None

  #region Test of normal operation.

  def test_plane(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Specify one Tk window in the file.
    """
    gen = Generator("tests/definition/generator_test/plane.yml")
    tk = gen.generate()
    self.assertEquals(type(tk), tkinter.Tk)
    self.assertEquals(tk.title(), "TEST Window")

  def test_button(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    """
    gen = Generator("tests/definition/generator_test/button.yml")
    tk = gen.generate()
    self.assertEquals(type(tk.children["!frame"]), tkinter.Frame)
    self.assertEquals(type(tk.children["!frame"].children["!button"]), tkinter.Button)

  def test_localize(self):
    """
    When you call the `Generator#generate()` method with the following conditions
    Verify that the translation is done.
    * Specify one Tk window in the file.
    * Specify existing translation dictionary files.
    """
    gen = Generator("tests/definition/generator_test/localize_target.yml", localization_file="tests/definition/generator_test/localize_script.yml")
    gen.generate()
    self.assertEquals(gen.findbyid("button").widget.cget("text"), "translated")

  def test_idtags(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * ID and tag elements are defined in windows and widgets.
    """
    gen = Generator("tests/definition/generator_test/idtags.yml")
    tk = gen.generate()
    self.assertEquals(type(gen.findbyid("test").widget), tkinter.Tk)
    self.assertEquals(type(gen.findbyid("testbutton").widget), tkinter.Button)
    self.assertIsNone(gen.findbyid("unknown"))

  def test_variable(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * Widget variables are set in the widget.
    """
    gen = Generator("tests/definition/generator_test/variable.yml")
    tk = gen.generate()
    self.assertIsNotNone(type(gen.findbyid("test1").widget["textvariable"]))
    self.assertIsNotNone(type(gen.findbyid("test2").widget["textvariable"]))
    self.assertIsNotNone(type(gen.findbyid("test3").widget["textvariable"]))
    self.assertIsNotNone(type(gen.findbyid("test4").widget["textvariable"]))
    self.assertIs(type(gen.vars["test1"]), tkinter.StringVar)

  def test_variable_in_array(self):
    """
    When the `Generator#generate()` method is called with the following conditions
    Make sure that all variable objects set in the parameters are expanded.
    * A var variable exists on a node other than the node directly under the widget.
    """
    gen = Generator(file="tests/definition/generator_test/variable_in_array.yml", modules=["tests.test_generator", "tkinter"])
    gen.generate()
    self.assertIs(type(gen.findbyid("button1").widget.items[0]), tkinter.IntVar)
    self.assertIs(type(gen.findbyid("button2").widget.testdict["testvar"]), tkinter.StringVar)

  def test_include(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Include another YAML file within a YAML file.
    """
    gen = Generator("tests/definition/generator_test/multiple_files.yml")
    tk = gen.generate()
    self.assertEquals(type(gen.findbyid("testbutton").widget), tkinter.Button)

  def test_widget_override(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Multiple modules are referenced.
    * Overriding an existing tkinter widget in the referenced module.
    """
    gen = Generator(file="tests/definition/generator_test/button.yml", modules=["tests.test_generator", "tkinter"])
    tk = gen.generate()
    self.assertEquals(gen.findbyid("test").widget.__class__.__module__, "tests.test_generator")

  def test_load_utf8(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Contains Japanese characters.
    * The encoding of the file is UTF-8.
    """
    gen = Generator("tests/definition/generator_test/japanese_utf8.yml")
    gen.generate()
    self.assertEquals(gen.findbyid("test").widget.cget("text"), "ボタン")

  def test_load_sjis(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Contains Japanese characters.
    * The encoding of the file is Shift-JIS.
    * Specify Shift-JIS as the encoding in the Generator argument.
    """
    gen = Generator("tests/definition/generator_test/japanese_sjis.yml", encoding="Shift_JIS")
    gen.generate()
    self.assertEquals(gen.findbyid("test").widget.cget("text"), "ボタン")

  #endregion

  #region Test of semi-normal operation

  def test_include_multidir(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * Include another YAML file within a YAML file.
    * The file path of the YAML file read the first time and the YAML file read the second time are different.
    """
    Generator("tests/definition/generator_test/multiple_files.yml")
    gen = Generator("tests/definition/generator_test/testdir/multiple_files.yml")
    tk = gen.generate()
    self.assertEquals(type(gen.findbyid("testbutton").widget), tkinter.Button)

  def test_no_localize(self):
    """
    When you call the `Generator#generate()` method with the following conditions
    Verify that no translation is done.
    * Specify one Tk window in the file.
    * Do not specify the translation dictionary file.
    """
    gen = Generator("tests/definition/generator_test/localize_target.yml")
    gen.generate()
    self.assertEquals(gen.findbyid("button").widget.cget("text"), "OK")

  def test_command_no_handler(self):
    """
    Confirm that the target Tk window is created when the `Generator#generate()` method
    is called under the following conditions.
    * There is a widget in the YAML file that contains the `::command` keyword
    * Argument command of `Generator#generate()` is not specified
    """
    gen = Generator("tests/definition/generator_test/has_command.yml")
    tk = gen.generate()
    self.assertTrue(True)

  #endregion

  #region Abnormal behavior test

  def test_variable_noname(self):
    """
    Confirm that ValueError occurs when calling the `Generator#generate()` method under the following conditions.
    * Specify one Tk window in the file.
    * The widget variable does not contain a name.
    """
    gen = Generator("tests/definition/generator_test/variable_error2.yml")
    with self.assertRaises(ValueError):
      gen.generate()

  def test_variable_no_variable(self):
    """
    Confirm that ValueError occurs when calling the `Generator#generate()` method under the following conditions.
    * Specify one Tk window in the file.
    * Specify the class name of a class that is not a subclass of Variable as the widget variable type.
    """
    gen = Generator("tests/definition/generator_test/variable_error3.yml")
    with self.assertRaises(ValueError):
      gen.generate()

  def test_variable_unknown_name(self):
    """
    Confirm that AttributeError occurs when calling the `Generator#generate()` method under the following conditions.
    * Specify one Tk window in the file.
    * The widget variable type is a non-existent class name.
    """
    gen = Generator("tests/definition/generator_test/variable_error4.yml")
    with self.assertRaises(AttributeError):
      gen.generate()

  def test_unkown_keyword(self):
    """
    Confirm that NameError occurs when calling the `Generator#generate()` method under the following conditions.
    * Specify one Tk window in the file.
    * Keyword is a non-existent name.
    """
    gen = Generator("tests/definition/generator_test/command_unknown.yml")
    with self.assertRaises(NameError):
      gen.generate()

  def test_load_sjis_noencode(self):
    """
    Confirm that UnicodeDecodeError occurs when calling the `Generator#generate()` method under the following conditions.
    * Contains Japanese characters.
    * The encoding of the file is Shift-JIS.
    * Do not specify the encoding in the Generator argument.
    """
    with self.assertRaises(UnicodeDecodeError):
      Generator("tests/definition/generator_test/japanese_sjis.yml")

  #endregion

if __name__ == "__main__":
  unittest.main()


import unittest

import yaml

from tksugar.localizer import Localizer

class Test_Localizer(unittest.TestCase):
  """
  Tests the `Localizer` Class
  """

  #region Testing for normal operation

  def test_prepare(self):
    """
    Confirm that the translation data can be read correctly when `Localizer#_prepare()` is executed under the following conditions.
    * YML file exists.
    * YML file has a tree structure
    """
    l = Localizer("tests/definition/localizer_test/safecase.yml")
    l._prepare()
    self.assertEqual(len(l._translatedict), 4)
    self.assertEqual(l._translatedict["testa"], "a")
    self.assertEqual(l._translatedict["test.testb"], "b")
    self.assertEqual(l._translatedict["test.testc"], "c")
    self.assertEqual(l._translatedict["test.test.testd"], "d")

  def test_translate(self):
    """
    If you run `Localizer#_translate()` under the following conditions,
    make sure that the string specified as an argument will be replaced by a string in the dictionary.
    * After executing `Localizer#_prepare()`.
    * The keywords present in the dictionary.
    """
    l = Localizer("tests/definition/localizer_test/safecase.yml")
    l._prepare()
    self.assertEqual(l._translate("testa"), "a")
    self.assertEqual(l._translate("test.testb"), "b")

  def test_haslist(self):
    """
    If you run `Localizer#_translate()` under the following conditions,
    Make sure that the items in the list are not stored.
    * YML file exists.
    * YML file contains list information.
    """
    l = Localizer("tests/definition/localizer_test/haslist.yml")
    l._prepare()
    self.assertEqual(l._translatedict["test.testc"], "")

  def test_localize(self):
    """
    If you run `Localizer#localize()` under the following conditions,
    Make sure that all keywords among the given data are replaced.
    * YML file exists.
    * A mixture of keywords in the dictionary and those not in the dictionary.
    """
    l = Localizer("tests/definition/localizer_test/safecase.yml")
    with open("tests/definition/localizer_test/target.yml") as f:
      data = yaml.safe_load(f)
    l.localize(data)
    self.assertEqual(data["targets"], "a")
    self.assertEqual(data["non_targets"], "no translate")
    self.assertEqual(data["listitems"], ["b", "c", "test.unknown"])
    self.assertEqual(data["longtext"], "test\nd\na")

  def test_localize_utf8(self):
    """
    If you run `Localizer#localize()` under the following conditions,
    Make sure that all keywords among the given data are replaced.
    * YML file exists.
    * A mixture of keywords in the dictionary and those not in the dictionary.
    * YML file contains Japanese characters.
    * The encoding of the file is UTF-8.
    """
    l = Localizer("tests/definition/localizer_test/japanese_utf8.yml")
    with open("tests/definition/localizer_test/target.yml") as f:
      data = yaml.safe_load(f)
    l.localize(data)
    self.assertEqual(data["targets"], "テストA")

  def test_localize_sjis(self):
    """
    If you run `Localizer#localize()` under the following conditions,
    Make sure that all keywords among the given data are replaced.
    * YML file exists.
    * A mixture of keywords in the dictionary and those not in the dictionary.
    * YML file contains Japanese characters.
    * The encoding of the file is Shift-JIS.
    * Specify Shift-JIS as the encoding in the Generator argument.
    """
    l = Localizer("tests/definition/localizer_test/japanese_sjis.yml", encoding="Shift_JIS")
    with open("tests/definition/localizer_test/target.yml") as f:
      data = yaml.safe_load(f)
    l.localize(data)
    self.assertEqual(data["targets"], "テストA")

  #endregion

  #region Semi-normal behavior testing

  def test_translate_nodict(self):
    """
    If you run `Localizer#_translate()` under the following conditions,
    make sure that the string specified as an argument not be replaced.
    * After executing `Localizer#_prepare()`.
    * The keywords not present in the dictionary.
    """
    l = Localizer("tests/definition/localizer_test/safecase.yml")
    l._prepare()
    self.assertEqual(l._translate("testb"), "testb")
    self.assertEqual(l._translate("test.testd"), "test.testd")

  def test_localize_emptytext(self):
    """
    If you run `Localizer#localize()` under the following conditions,
    Ensure that no exceptions are raised and all keywords in the specified data are not replaced.
    * Specify blanks at the specified location in the YML file.
    * A mixture of keywords in the dictionary and those not in the dictionary.
    """
    l = Localizer("")
    with open("tests/definition/localizer_test/target.yml") as f:
      data = yaml.safe_load(f)
    l.localize(data)
    self.assertEqual(data["targets"], "testa")
    self.assertEqual(data["non_targets"], "no translate")
    self.assertEqual(data["listitems"], ["test.testb", "test.testc", "test.unknown"])
    self.assertEqual(data["longtext"], "test\ntest.test.testd\ntesta")

  def test_localize_nofile(self):
    """
    If you run `Localizer#localize()` under the following conditions,
    Ensure that no exceptions are raised and all keywords in the specified data are not replaced.
    * YML file not exists.
    * A mixture of keywords in the dictionary and those not in the dictionary.
    """
    l = Localizer("tests/definition/localizer_test/unknown.yml")
    with open("tests/definition/localizer_test/target.yml") as f:
      data = yaml.safe_load(f)
    l.localize(data)
    self.assertEqual(data["targets"], "testa")
    self.assertEqual(data["non_targets"], "no translate")
    self.assertEqual(data["listitems"], ["test.testb", "test.testc", "test.unknown"])
    self.assertEqual(data["longtext"], "test\ntest.test.testd\ntesta")

  #endregion

  #region Anomaly Testing

  def test_localize_sjis_noencode(self):
    """
    If you run `Localizer#localize()` under the following conditions,
    Make sure that all keywords among the given data are replaced.
    * YML file exists.
    * A mixture of keywords in the dictionary and those not in the dictionary.
    * YML file contains Japanese characters.
    * The encoding of the file is Shift-JIS.
    * The encoding of the file is Shift-JIS.
    * Do not specify the encoding in the Generator argument.
    """
    with self.assertRaises(UnicodeDecodeError):
      Localizer("tests/definition/localizer_test/japanese_sjis.yml")

  #endregion

if __name__ == "__main__":
  unittest.main()


import importlib
import inspect
from pathlib import Path
import tkinter
import re
from typing import Type

import yaml
from yamlinclude import YamlIncludeConstructor

from tksugar.tkmanager import TkManager
from tksugar.localizer import Localizer
from tksugar.widgets.generatorsupport import GeneratorSupport
from tksugar.eventreciever import EventReciever

class TagData(object):
  """
  An object that represents additional data for the widget.
  In TkManager, it is used to link the TkManager ID and the widget.
  """
  def __init__(self, widget):
    """
    Constructor

    Parameters
    ----
    widget: object
      Tk widget.
    """
    self.widget = widget
    self.id = None
    self.tag= None
    self.callback = None

  def hasdata(self):
    """
    True if there is data.

    Returns
    ----
    hasdata: bool
      True if there is data
    """
    return self.id or self.tag

  def performclick(self):
    """
    If the button has a command, execute it.
    """
    if self.callback is not None:
      self.callback(self.widget, self)

class TemporaryVariable(object):
  """
  A temporary variable object that indicates where to replace the tkinter.Variable object.
  Hold only the variable name.
  """
  def __init__(self, name):
    self.name = name

class GeneratorLoader(yaml.SafeLoader):
  """
  YAML Loader used in Generator.
  A custom tag reading process is added.
  """
  def __init__(self, stream):
    super().__init__(stream)
    self.vars = {}
    yaml.add_multi_constructor("tag:yaml.org,2002:var", GeneratorLoader.var_handler,
      Loader=GeneratorLoader)

  @staticmethod
  def var_handler(loader, suffix, node=None):
    """
    A handler that responds to variable definitions.
    """
    if suffix[0] == ":": suffix = suffix[1:]
    name = ""
    default = None
    for v in node.value:
      if v[0].value == "name": name = v[1].value
      if v[0].value == "default": default = v[1].value
    if name != "":
      var = getattr(tkinter, suffix)
      if not issubclass(var, tkinter.Variable): raise ValueError("The specified class is not a Variable class.")
      loader.vars[name] = {"class": var, "default": default}
      return TemporaryVariable(name)
    else:
      raise ValueError("The variable name is not set.")

#region command classes

class CommandBaseClass(object):
  """
  A base class that defines the content of commands in YAML files.
  """
  def __init__(self, postactions):
    """
    constuructor

    Parameters
    ----
    postactions: list
      If there is a method you want to call after creating the object list, add it to this list.
    """
    self.postactions = postactions

  def __call__(self, obj, tag, value):
    self.command(obj, tag, value)

  def command(self, obj, tag, value):
    """
    Command executor

    Parameters
    ----
    obj: object
      The associated object.
    tag: Any
      Tag object.
    value: Any
      The value defined in the YAML file.
    """
    raise NotImplementedError

class IdCommand(CommandBaseClass):
  """
  A command that associates an object with an internal ID.
  """
  def command(self, object, tag, value):
    tag.id = str(value)

class TagCommand(CommandBaseClass):
  """
  A command that associates an object with tag data.
  """
  def command(self, object, tag, value):
    tag.tag = value

class CommandCommand(CommandBaseClass):
  """
  A command that associates a callback method that responds to an object's command.
  """
  def __init__(self, postactions, callback):
    """
    Constructor

    Parameters
    ----
    callback: func
      An event handler for processing commands for widgets with the ::command element set.
    """
    super().__init__(postactions)
    self.callback = callback

  def command(self, object, tag, value):
    try:
      if not self.callback is None:
        resv = EventReciever(object, tag, self.callback)
        if "command" in dir(object):
          setattr(object, "command", resv)
        else:
          object["command"] = resv
    except tkinter.TclError:
      pass

class GridColumnCommand(CommandBaseClass):
  """
  Configure columns on the grid.

  Parameters at Commands
  ----
  index: int
    Column index.
  minsize: int
    Specify the minimum width / height numerically.
  weight: int
    Specify the ratio when allocating the margin numerically.
  pad: int
    Specify the padding numerically.
  """

  def command(self, object, tag, value):
    def _command():
      indx = value["index"]
      del value["index"]
      object.grid_columnconfigure(indx, value)

    self.postactions.append(_command)

class GridRowCommand(CommandBaseClass):
  """
  Configure columns on the grid.

  Parameters at Commands
  ----
  index: int
    Row index.
  minsize: int
    Specify the minimum width / height numerically.
  weight: int
    Specify the ratio when allocating the margin numerically.
  pad: int
    Specify the padding numerically.
  """

  def command(self, object, tag, value):
    def _command():
      indx = value["index"]
      del value["index"]
      object.grid_rowconfigure(indx, value)

    self.postactions.append(_command)

#endregion

class Generator(object):
  """
  The core object that creates the Tk window.
  Users of this module will use this core object to generate a Tk window.
  """
  def __init__(self, file="",modules=["tksugar.widgets", "tkinter"], localization_file="", encoding="UTF-8"):
    """
    constructor.

    Parameters
    ----
    file: str
      A file path describing the window's object and layout.
      This argument can be omitted, but in actual use it is not omitted in principle.
      Omitted only when testing.
      If you omit the file name, YamlIncludeConstructor will not be deleted, so YAML's `!Include` will remain valid.
      However, this operation is not dealt with because it is impossible in principle to omit the file in the first place and there is no actual harm.
    modules: list[str]
      An array indicating the name of the module to be used.
      By default, it is "tkinter" only.

      Because modules can be added with the `Generator#add_modules()`,
      the value is specified here only if you do not want to load the tkinter module.
    localization_file: str
      Path indicating a YAML-formatted dictionary file used for UI localization.
      This dictionary contains only dict, list, and str.
      If the format is wrong, no error is generated at this point, even if the path does not exist.

      If omitted, the UI localizer is disabled.
    encoding: str
      File Encoding.
    """
    self.string = ""
    if file:
      with open(file, "r", encoding=encoding) as f:
        self.string = f.read()
      YamlIncludeConstructor.add_to_loader_class(loader_class=GeneratorLoader, base_dir=str(Path(file).parent))
    self._modules = modules
    self._widgets = []
    self.localization_file = localization_file
    self.localization_file_encoding = encoding
    self.vars = None

  def add_modules(self, *modules):
    """
    Add a module to be used.

    Parameters
    ----
    modules: list[str]
      module names.
    """
    self._modules.append(*modules)

  def generate(self, command=None):
    """
    Generate a Tk window based on the specified files and modules.

    Parameters
    ----
    command: func
      An event handler for processing commands for widgets with the ::command element set.

    Returns
    ----
    window: tkinter.Tk
      Tk window object.
    """
    def _generate_core(children, owner, modules):
      for i in children:
        cls = self._load_class(modules, i["classname"])
        objparam = i["params"]
        if not issubclass(type(owner), GeneratorSupport):
          # GenetratorSupport non inherited class, which adds a master parameter and adds a child object.
          objparam["master"] = owner
        obj, tag = self._instantiate(cls, callback=command, **objparam)
        if issubclass(type(owner), GeneratorSupport):
          # GeneratorSupport inherited class, which adds a child object via append_child.
          childparam = {}
          for n, v in objparam.items():
            if n.startswith("/"):
              childparam[n[1:]] = v
          owner.append_child(obj, **childparam)
        if tag.hasdata(): self._widgets.append(tag)
        if i["children"]:
          _generate_core(i["children"], obj, modules)
    # Load YAML
    loader = GeneratorLoader(self.string)
    try:
      struct = loader.get_single_data()
    finally:
      loader.dispose()
    self.vars = loader.vars
    if not type(struct) is dict or len(struct) > 1:
      raise ValueError("The root node must be a dict and single.")
    # Prepare
    l = Localizer(self.localization_file, self.localization_file_encoding)
    l.localize(struct)
    modules = self._load_modules()
    tree = self._scantree(struct)
    # Load Root Object
    cls  = self._load_class(modules, tree["classname"])
    root, tag = self._instantiate(cls, callback=command, **tree["params"])
    if tag.hasdata(): self._widgets.append(tag)
    # Load Variable
    for n, v in self.vars.items():
      self.vars[n] = v["class"](master=root, name=n)
      if not v["default"] is None:
        self.vars[n].set(v["default"])
    # Load Child Object
    _generate_core(tree["children"], root, modules)
    return root

  def findbyid(self, id):
    """
    Search for widgets by ID.
    If multiple items with the same ID are defined, the first item is returned.

    Parameters
    ----
    id: str
      ID

    Returns
    ----
    widget: TagData|None
      If an item is found, TagData containing that widget.
      None if the item is not found.
    """
    l = list(filter(lambda x: x.id == id, self._widgets))
    return None if l == [] else l[0]

  def get_manager(self, commandhandler=None):
    """
    Create a window, store it in the `TkManager` that manages the window, and return it.

    Returns
    ----
    manager: TkManager
      A TkManager object that contains a window object.
    commandhandler: func
      An event handler for processing commands for widgets with the ::command element set.
    """
    window = self.generate(command=commandhandler)
    return TkManager(window, self._widgets, self.vars)

  ### Private Methods

  def _load_modules(self):
    """
    Read all modules specified in the `self._modules` array
    and return the array.

    Returns
    ----
    modules: dict[str, module]
      A dictionary object that associates module names with module objects.

    Raises
    ----
    ModuleNotFoundError
      There is no module with the specified name.
    """
    modules = {}
    for mod in self._modules:
      modules[mod] = importlib.import_module(str(mod))
    return modules

  @staticmethod
  def _load_class(modules, class_name):
    """
    Search the class with the specified name and
    return the class object of the found class.

    Parameters
    ----
    modules: dict[str, module]
      A dictionary object that associates module names with module objects.
    class_name: str
      name of the class.
      If the class name contains ".", the left side of "."
      Is regarded as the module name and only the module with that name is searched.
      If the beginning of the character string is "_", it is ignored and processed.

    Returns
    ----
    class_object: class
      Class object.

    Raises
    ----
    TypeError
      If the specified class does not exist in the module specified by the argument.
      Or, if the specified class does not exist in the specified module.
    """
    mod = None
    cls = None
    if class_name[0] == "_":
      class_name = class_name[1:]
    if "." in class_name:
      # Module specified
      mod, cls = class_name.rsplit(".", 1)
    else:
      # Module search
      cls = class_name
      for module in modules.keys():
        cllist = list(map(lambda x: x[0], inspect.getmembers(modules[module], inspect.isclass)))
        if cls in cllist:
          mod = module
          break

    if mod is None:
      raise TypeError(f'Class not found. "{class_name}"')
    try:
      return getattr(modules[mod], cls)
    except AttributeError as e:
      raise TypeError(e)

  @staticmethod
  def _scantree(struct):
    """
    Scan an array and convert it to a tree of class names, parameters and child objects

    Parameters
    ----
    struct: dict
      Data array

    Retrns
    ----
    treedata: dict
      Tree data
    """
    def _scantree_core(struct, params):
      props = {
        "classname": "",
        "params": {},
        "children": [],
      }
      rootname = next(iter(struct))
      props["classname"] = rootname[1:]
      items = {}
      # merge params.
      if "params" in params:
        if struct[rootname] is None:
          items = params["params"]
        else:
          items = dict(params["params"], **struct[rootname])
      else:
        items = struct[rootname]
        if items is None:
          raise ValueError(f'Missing value or element in node name "{rootname}". Is the indentation level wrong?')
      # rootname check.
      if rootname == "::params":
        params["params"] = items
        return None
      # parse.
      for n, v in items.items():
        if n[0] == "_":
          props["children"].append(_scantree_core({n: v}, params))
        elif n == "::children":
          if type(v) is not list:
            raise AttributeError("The child elements of the ::children node must be an list.")
          inparam = dict(params)
          for item in v:
            r = _scantree_core(item, inparam)
            if r is not None: props["children"].append(r)
        elif n == "::params":
          params["params"] = v
        else:
          if v is None:
            props["params"][n] = None
          else:
            props["params"][n] = v
      return props
    return _scantree_core(struct, {})

  @staticmethod
  def _get_argnames(method):
    """
    Get method argument list.
    If the document comment includes "STANDARD OPTIONS" and "WIDGET-SPECIFIC" OPTIONS,
    use that as the argument list.

    Parameters
    ----
    method: func
      Function object.

    Returns
    ----
    arglist: list(str)
      Argument list.
    """
    result = []
    # add inspect result
    for p in filter(lambda p: p.kind == p.POSITIONAL_OR_KEYWORD, inspect.signature(method).parameters.values()):
      result.append(p.name)

    # add comment args
    lines = method.__doc__.split("\n") if method.__doc__ else []
    i = 0
    collect = False
    while i < len(lines):
      if collect:
        if lines[i] == "":
          collect = False
        else:
          for p in filter(lambda x: x, re.split("[\s,]", lines[i].strip())):
            result.append(p.strip())
      else:
        if "STANDARD OPTIONS" in lines[i] or "WIDGET-SPECIFIC OPTIONS" in lines[i]:
          collect = True
          i += 1
        if "Valid resource names:" in lines[i]:
          collect = True
          lines[i] = lines[i].split(":")[1]
          i -= 1
      i += 1
    return result

  @staticmethod
  def _split_params(method, params):
    """
    Separate the parameter list into optional method arguments and anything else.

    Parameters
    ----
    method: func
      Method to be distributed.
    params: dict[str, any]
      Parameter list.

    Returns
    ----
    methodparams: dict[str, any]
      method arguments.
    other: dict[str, any]
      Any other value.
    """
    methodparams = {}
    for p in Generator._get_argnames(method):
      if p in params:
        methodparams[p] = params[p]
        del params[p]
    return methodparams, params

  def _instantiate(self, cls, callback=None, **params):
    """
    Generate an object with set properties based on class and property list.

    Parameters
    ----
    cls: class
      The object to instantiate.
    callback: func
      An event handler for processing commands for widgets with the ::command element set.
    params: dict(str, any)
      Property list.

    Returns
    ----
    instance: object
      The instantiated class.
    tagdata: TagData
      Widget additional data.
    """
    def replace_variable(params):
      """
      Recursively replaces the TemporaryVariable class present in all parameters.
      """
      for n in params.keys() if type(params) is dict else range(len(params)):
        if type(params[n]) is TemporaryVariable:
          params[n] = self.vars[params[n].name]
        elif type(params[n]) is list or type(params[n]) is dict:
          replace_variable(params[n])
    # Prepare
    replace_variable(params)
    initparams, others = Generator._split_params(cls.__init__, params)
    postactions = []
    commands = {
      "id": IdCommand(postactions),
      "tag": TagCommand(postactions),
      "gridcolumn": GridColumnCommand(postactions),
      "gridrow": GridRowCommand(postactions),
    }
    commands["command"] = CommandCommand(postactions, callback)
    # Instantiation
    obj = cls(**initparams)
    tagdata = TagData(obj)
    if callback is not None:
      tagdata.callback = callback
    # Other property settings
    for n, v in others.items():
      if n.startswith("::"):
        if n[2:] in commands:
          commands[n[2:]](obj, tagdata, v)
        else:
          raise NameError("Command Not Found('{0}')".format(n[2:]))
      elif n.startswith("/"):
        pass
      elif inspect.isroutine(getattr(obj, n)):
        attr = getattr(obj, n)
        attr() if v is None else attr(v)
      else:
        setattr(obj, n, v)
    # Post actions
    for fun in postactions:
      fun()
    return obj, tagdata

if __name__ == "__main__":
  gen = Generator()
  gen.string = """
  _Tk:
    title: "TEST Window"
    geometry: 400x300
    _Frame:
      pack:
      _Label:
        text: "Hello"
        pack:
  """
  window = gen.generate()
  window.mainloop()


from pathlib import Path
import re

from yaml.loader import SafeLoader

class Localizer(object):
  """
  Translates all text of the specified data based on the dictionary data for translation.
  The string to be translated in the data list is decorated with three colons (:::) (ex. `:::text`)

  The operation of this class is to replace the character string to be translated with the text in the dictionary data.

  Dictionary data can have a hierarchical structure.
  Data with a hierarchical structure is expanded into a character string delimited during translation processing.
  """
  def __init__(self, file, encoding="UTF-8"):
    """
    Constructor.

    Parameters
    ----
    file: str
      The path to the YAML file that contains the translation string.
    encoding: str
      File Encoding.
    """
    self.string = ""
    p = Path(file)
    if file and p.exists():
      with open(p, "r", encoding=encoding) as f:
        self.string = f.read()
    self._translatedict = None

  def _prepare(self):
    """
    Read and prepare the data list.
    """
    def flatten_dict(basename, struct):
      result = {}
      if type(struct) is dict:
        for k in struct.keys():
          if type(struct[k]) is dict:
            result.update(flatten_dict(f"{basename}{k}.", struct[k]))
          elif type(struct[k]) is list:
            result[f"{basename}{k}"] = ""
          else:
            result[f"{basename}{k}"] = struct[k]
      return result
    if self._translatedict is None:
      # Read
      loader = SafeLoader(self.string)
      try:
        data = loader.get_single_data()
      finally:
        loader.dispose()
      self._translatedict = flatten_dict("", data)

  def localize(self, data):
    """
    Translate the given data.
    This method modifies the given data directly.

    Parameters
    ----
    data: dict
      Data list.
    """
    def translate_core(data):
      for k in data.keys() if type(data) is dict else range(len(data)):
        if type(data[k]) is dict:
          translate_core(data[k])
        elif type(data[k]) is list:
          translate_core(data[k])
        elif type(data[k]) is str:
          data[k] = rexp.sub(lambda m: self._translate(m.group(0)[3:]), data[k])
    rexp = re.compile(r":::\S+")
    self._prepare()
    translate_core(data)

  def _translate(self, name):
    """
    Perform translation processing.

    Parameters
    ----
    name: str
      Keyword name.

    Returns
    ----
    return: str
      The string after replacement.
      If the keywords were not present in the dictionary, the string is returned as is without replacement.
    """
    return self._translatedict[name] if name in self._translatedict else name



import tkinter.ttk


from tksugar.widgets.generatorsupport import GeneratorSupport

class Notebook(tkinter.ttk.Notebook, GeneratorSupport):
  """
  `tkinter.ttk.Notebook` with added methods to support Generator.
  """
  def append_child(self, child, **params):
    self.add(child, **params)


from tksugar.eventreciever import EventReciever
class TkManager(object):
  """
  Manager object for managing widgets generated by the `tksugar.Generator` object.
  Manages IDs and event handlers, and manages variables.
  """
  def __init__(self, window, widgets, vars):
    """
    Constructor

    Parameters
    ----
    window: Tk
      Tk window object
    widgets: list[TagData]
      An array of TagData objects containing widgets with ids.
    vars: dict[str, Variable]
      A dictionary containing widget variables declared by tags.
    """
    self._window = window
    self.widgets = {}
    self.vars = vars
    self.trace_handler = None
    for tagdata in widgets:
      tagdata.tag = {
        "tag": tagdata.tag
      }
      if not tagdata.id in self.widgets:
        self.widgets[tagdata.id] = tagdata

    for n, v in vars.items():
      v.trace("w", EventReciever(v, n, self._tracevars))

  def _tracevars(self, obj, name):
    if self.trace_handler:
      self.trace_handler(obj, name)

  def mainloop(self):
    """
    Call the window's main loop.
    """
    self._window.mainloop()

  @property
  def window(self):
    """
    Get a window object.
    """
    return self._window


from abc import abstractmethod

class GeneratorSupport(object):
  """
  An abstract class that defines methods to support your own custom processing.
  Generator compatible class inherits this class and implements necessary methods.
  """

  @abstractmethod
  def append_child(self, child, **params):
    """
    Called when storing a child object.

    Parameters
    ----
    child: tkinter.Widget
      A list of objects that are child elements of the object.
    params: dict
      Parameters.
    """
    raise NotImplementedError


"""
Widgets that require Generator-specific custom processing are stored in this package.
For example, widgets such as tkinter.ttk.Notebook that do not display child elements on the GUI just by referencing them with the master argument of child elements are recorded here.
This package is referenced with the highest priority when the Generator is executed.
Therefore, even if there is a widget with the same name in another module, this module will be called.
"""
from tksugar.widgets.generatorsupport import GeneratorSupport
from tksugar.widgets.notebook import Notebook
from tksugar.widgets.menu import Menu


from msilib.schema import Class
import unittest
import types
import tkinter

import yaml

from tksugar.generator import Generator

class ClassForTest(object):
  def __init__(self, a, b, c, d=1, e=2, f=3):
    self.a = a
    self.b = b
    self.c = c
    self.d = d
    self.e = e
    self.f = f
    self.g = None
    self.h = False
    self.i = None

  def seth(self):
    self.h = True

  def seti(self, value):
    self.i = value

  @staticmethod
  def j():
    """
    This method has documentation comments.
    STANDARD OPTIONS

    g, i,
    j, k, l
    """
    pass

  @staticmethod
  def k():
    """
    This method has documentation comments.

    Valid resource names: g, i,
    j, k, l
    """
    pass

class Test_Generator_Methods(unittest.TestCase):
  """
  Method tests other than the `generate()` method of the Generator class.
  """

  #region test of add_module()

  def test_add_module(self):
    """
    Confirm that the module added by `Generator#add_modules()`
    is treated the same as the module added by the constructor.
    """
    gen = Generator(modules=["tkinter"])
    gen.add_modules("tkinter.ttk")
    mods = gen._load_modules()
    self.assertEquals(len(mods), 2)
    self.assertEquals(type(mods["tkinter"]), types.ModuleType)
    self.assertEquals(type(mods["tkinter.ttk"]), types.ModuleType)

  #endregion

  #region test of _load_module()
  def test_load_module_empty(self):
    """
    Confirm that an empty array is returned when `Generator#_load_modules()` is
    called under the following conditions.
    * The number of modules specified in the _modules array is 0
    """
    gen = Generator(modules=[])
    mods = gen._load_modules()
    self.assertEquals(len(mods), 0)

  def test_load_module_module_not_found(self):
    """
    Confirm that ModuleNotFoundError is thrown
    when `Generator#_load_modules()` is called under the following conditions.
    * The module with the specified name does not exist.
    """
    gen = Generator(modules=["unknown_module"])
    with self.assertRaises(ModuleNotFoundError):
      gen._load_modules()

  def test_load_module_module_not_found_at_once(self):
    """
    Confirm that ModuleNotFoundError is thrown
    when `Generator#_load_modules()` is called under the following conditions.
    * Modules that exist and modules that do not exist are mixed in the module name array.
    """
    gen = Generator(modules=["tkinter", "unknown_module"])
    with self.assertRaises(ModuleNotFoundError):
      gen._load_modules()

  def test_load_module_empty_string(self):
    """
    Confirm that ValueError is thrown
    when `Generator#_load_modules()` is called under the following conditions.
    * The array contains an empty string.
    """
    gen = Generator(modules=[""])
    with self.assertRaises(ValueError):
      gen._load_modules()

  def test_load_module_simple(self):
    """
    Make sure that `Generator#_load_modules()` loads all modules under the following conditions.
    * Only one module specified in the argument.
    """
    gen = Generator(modules=["tkinter"])
    mods = gen._load_modules()
    self.assertEquals(len(mods), 1)
    self.assertEquals(next(iter(mods)), "tkinter")

  def test_load_module_multiple_items(self):
    """
    Make sure that `Generator#_load_modules()` loads all modules under the following conditions.
    * Multiple module names are specified in the argument.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    self.assertEquals(len(mods), 2)
    self.assertEquals(type(mods["tkinter"]), types.ModuleType)
    self.assertEquals(type(mods["tkinter.ttk"]), types.ModuleType)

  #endregion

  #region test of _load_class()

  def test_load_class_anymodule_oneclass(self):
    """
    Confirm that the class is searched
    when `Generator#_load_class()` is called under the following conditions.
    * No module name specified.
    * The class exists in the loaded module list.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    cls = Generator._load_class(mods, "Notebook")
    self.assertEquals(type(cls), type)

  def test_load_class_onemodule_oneclass(self):
    """
    Confirm that the class is searched
    when `Generator#_load_class()` is called under the following conditions.
    * Module name is specified.
    * The class exists in the loaded module list.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    cls = Generator._load_class(mods, "tkinter.ttk.Notebook")
    self.assertEquals(type(cls), type)

  def test_load_class_underscore(self):
    """
    Confirm that the class is searched
    when `Generator#_load_class()` is called under the following conditions.
    * There is an underscore before the module name.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    cls = Generator._load_class(mods, "_Notebook")
    self.assertEquals(type(cls), type)

  def test_load_class_unmatched_anymodule_oneclass(self):
    """
    Confirm that a TypeError occurs when calling `Generator#_load_class()`
    under the following conditions.
    * No module name specified.
    * Class does not exist in the loaded module list.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    with self.assertRaises(TypeError):
      gen._load_class(mods, "UnknownType")

  def test_load_class_unmatched_onemodule_oneclass(self):
    """
    Confirm that a TypeError occurs when calling `Generator#_load_class()`
    under the following conditions.
    * Module name is specified.
    * Class does not exist in the loaded module list.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    with self.assertRaises(TypeError):
      gen._load_class(mods, "tkinter.UnknownType")

  def test_load_class_different_module_onemodule_oneclass(self):
    """
    Confirm that a TypeError occurs when calling `Generator#_load_class()`
    under the following conditions.
    * Module name is specified.
    * The class exists in the loaded module list but does not exist in the specified module.
    """
    gen = Generator(modules=["tkinter", "tkinter.ttk"])
    mods = gen._load_modules()
    with self.assertRaises(TypeError):
      gen._load_class(mods, "tkinter.Notebook")

  #endregion

  #region test of _scantree()

  def test_scantree_simpledata(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    make sure that the method structures the tree.
    * Specify one Tk window in the file.
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/plane.yml", "r") as f:
      struct = yaml.safe_load(f)
      tree = gen._scantree(struct)
      self.assertEqual(tree["classname"], "Tk")
      self.assertEqual(len(tree["params"]), 1)
      self.assertEqual(tree["params"]["title"], "TEST Window")
      self.assertEqual(len(tree["children"]), 0)

  def test_scantree_button(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    make sure that the method structures the tree.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/button.yml", "r") as f:
      struct = yaml.safe_load(f)
      tree = gen._scantree(struct)
      self.assertEqual(tree["classname"], "Tk")
      self.assertEqual(tree["children"][0]["classname"], "Frame")
      self.assertEqual(tree["children"][0]["params"]["grid"], None)
      self.assertEqual(tree["children"][0]["children"][0]["classname"], "Button")
      self.assertEqual(tree["children"][0]["children"][0]["params"]["anchor"], "s")
      self.assertEqual(tree["children"][0]["children"][0]["params"]["text"], "OK")

  def test_scantree_children(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    make sure that the method structures the tree.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * `::children` element exists.
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/children.yml", "r") as f:
      struct = yaml.safe_load(f)
      tree = gen._scantree(struct)
      self.assertEqual(tree["classname"], "Tk")
      self.assertEqual(len(tree["children"][0]["children"]), 2)
      self.assertEqual(tree["children"][0]["children"][0]["classname"], "Label")
      self.assertEqual(tree["children"][0]["children"][1]["classname"], "Label")

  def test_scantree_children_notlist(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    Confirm that AttributeError occurs.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * `::children` element exists.
    * The child element of the `::children` element is not a list(dict).
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/children_error.yml", "r") as f:
      struct = yaml.safe_load(f)
      with self.assertRaises(AttributeError):
        gen._scantree(struct)

  def test_scantree_params_inside_child(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    That the method is structuring the tree,
    And, make sure that the value of `::params` affects only the elements under the `::children` element.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * The `::params` element exists inside the `::children` element.
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/params1.yml", "r") as f:
      struct = yaml.safe_load(f)
      tree = gen._scantree(struct)
      self.assertEqual(tree["classname"], "Tk")
      self.assertEqual(tree["children"][0]["children"][0]["params"]["text"], "Hello")
      self.assertEqual(tree["children"][0]["children"][1]["params"]["text"], "Test")
      self.assertEqual(tree["children"][0]["children"][2]["params"]["text"], "Test")
      self.assertNotIn("text", tree["children"][1]["params"])

  def test_scantree_params_outside_child(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    make sure that the method structures the tree.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * The `::params` element exists outside the `::children` element.
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/params2.yml", "r") as f:
      struct = yaml.safe_load(f)
      tree = gen._scantree(struct)
      self.assertEqual(tree["classname"], "Tk")
      self.assertIn("pack", tree["children"][0]["children"][0]["params"])
      self.assertIn("pack", tree["children"][0]["children"][1]["params"])

  def test_scantree_idtags(self):
    """
    When calling the `Generator#_scantree()` method under the following conditions,
    That the method is structuring the tree,
    And make sure to store the ID and tag in the tree without changing the values ​​of the ID and tag fields.
    * Specify one Tk window in the file.
    * There is a widget in the window.
    * ID and tag elements are defined in windows and widgets.
    """
    gen = Generator(modules=["tkinter"])
    with open("tests/definition/generator_test/idtags.yml", "r") as f:
      struct = yaml.safe_load(f)
      tree = gen._scantree(struct)
      self.assertEqual(tree["params"]["::id"], "test")
      self.assertEqual(tree["children"][0]["params"]["::tag"], "test")
      self.assertEqual(tree["children"][0]["children"][0]["params"]["::id"], "testbutton")
      self.assertEqual(tree["children"][0]["children"][0]["params"]["::tag"], "thisisatest")

  #endregion

  #region test of _get_argnames()

  def test_get_argnames(self):
    """
    When you call `Generator#get_argnames()` under the following conditions,
    Make sure to generate an argument list.
    * Common Python methods
    """
    list = Generator._get_argnames("".split)
    self.assertIn("maxsplit", list)

  def test_get_argnames_tkinterobj(self):
    """
    When you call `Generator#_get_argnames()` under the following conditions,
    make sure to add the contents of doc comment to the return value.
    * Target is tkinter module class __init__.
    """
    list = Generator._get_argnames(tkinter.Label.__init__)
    self.assertIn("activebackground", list)
    self.assertIn("text", list)
    self.assertIn("cnf", list)

  def test_get_argnames_hasnt_docstring(self):
    """
    When you call `Generator#get_argnames()` under the following conditions,
    Make sure to generate an argument list.
    * Document comment is not set in the method passed as an argument
    """
    list = Generator._get_argnames(ClassForTest.__init__)
    self.assertIn("a", list)
    self.assertIn("b", list)
    self.assertIn("c", list)

  def test_get_argnames_has_docstring(self):
    """
    When you call `Generator#get_argnames()` under the following conditions,
    Make sure to generate an argument list.
    * Document comment is set in the method passed as an argument
    """
    list = Generator._get_argnames(ClassForTest.j)
    self.assertListEqual(list, ["g", "i", "j", "k", "l"])

  def test_get_argnames_has_docstring_shorttype(self):
    """
    When you call `Generator#get_argnames()` under the following conditions,
    Make sure to generate an argument list.
    * Document comment is set in the method passed as an argument
    * The format of DocString is short.
    """
    list = Generator._get_argnames(ClassForTest.k)
    self.assertListEqual(list, ["g", "i", "j", "k", "l"])

  #endregion

  #region test of _split_params()

  def test_split_params(self):
    """
    When you call `Generator#_split_params()` under the following conditions,
    Confirm that the parameter list is returned correctly.
    * Target method parameters in the parameter list and other values ​​mixed.
    * All items in the parameter list have values.
    """
    params, others = Generator._split_params(tkinter.Label.__init__, {
      "text": "test",
      "bitmap": "abc",
      "test": "abc",
      "unknown": 123,
    })
    self.assertIn("text", params)
    self.assertIn("bitmap", params)
    self.assertIn("test", others)
    self.assertIn("unknown", others)

  def test_split_params_only_params(self):
    """
    When you call `Generator#_split_params()` under the following conditions,
    Confirm that the parameter list is returned correctly.
    * Only the parameter of the target method exists in the parameter list.
    * All items in the parameter list have values.
    """
    params, others = Generator._split_params(tkinter.Label.__init__, {
      "text": "test",
      "bitmap": "abc",
    })
    self.assertIn("text", params)
    self.assertIn("bitmap", params)
    self.assertEqual(len(others), 0)

  def test_split_params_only_others(self):
    """
    When you call `Generator#_split_params()` under the following conditions,
    Confirm that the parameter list is returned correctly.
    * Only the parameter of the target method exists in the parameter list.
    * All items in the parameter list have values.
    """
    params, others = Generator._split_params(tkinter.Label.__init__, {
      "test": "abc",
      "unknown": 123,
    })
    self.assertEqual(len(params), 0)
    self.assertIn("test", others)
    self.assertIn("unknown", others)

  def test_split_params_none_value(self):
    """
    When you call `Generator#_split_params()` under the following conditions,
    Confirm that the parameter list is returned correctly.
    * Target method parameters in the parameter list and other values ​​mixed.
    * Some of the items included in the parameter are None.
    """
    params, others = Generator._split_params(tkinter.Label.__init__, {
      "text": "test",
      "bitmap": None,
      "test": None,
      "unknown": 123,
    })
    self.assertIn("text", params)
    self.assertIn("bitmap", params)
    self.assertIn("test", others)
    self.assertIn("unknown", others)

  #endregion

  #region test of _instantiate()

  def test_instantiate(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Make sure the class is instantiated.
    * Specify all required parameters.
    """
    obj, unused = Generator()._instantiate(ClassForTest, a = "a", b = "b", c = "c")
    self.assertEqual(obj.a, "a")
    self.assertEqual(obj.b, "b")
    self.assertEqual(obj.c, "c")
    self.assertEqual(obj.d, 1)
    self.assertEqual(obj.e, 2)
    self.assertEqual(obj.f, 3)
    self.assertEqual(obj.g, None)
    self.assertEqual(obj.h, False)
    self.assertEqual(obj.i, None)

  def test_instantiate_non_mandatory_params(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Make sure the class is instantiated.
    * Specify all required parameters.
    * Specify all non-mandatory parameters.
    """
    obj, unused = Generator()._instantiate(ClassForTest, a = "a", b = "b", c = "c", d = "d", e = "e", f = "f")
    self.assertEqual(obj.a, "a")
    self.assertEqual(obj.b, "b")
    self.assertEqual(obj.c, "c")
    self.assertEqual(obj.d, "d")
    self.assertEqual(obj.e, "e")
    self.assertEqual(obj.f, "f")
    self.assertEqual(obj.g, None)
    self.assertEqual(obj.h, False)
    self.assertEqual(obj.i, None)

  def test_instantiate_set_other_props(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Make sure the class is instantiated.
    * Specify all required parameters.
    * Hold parameters that are not in constructor arguments.
    """
    obj, unused = Generator()._instantiate(ClassForTest, a = "a", b = "b", c = "c", g = "g", seth = None, seti = "abc")
    self.assertEqual(obj.a, "a")
    self.assertEqual(obj.b, "b")
    self.assertEqual(obj.c, "c")
    self.assertEqual(obj.d, 1)
    self.assertEqual(obj.e, 2)
    self.assertEqual(obj.f, 3)
    self.assertEqual(obj.g, "g")
    self.assertEqual(obj.h, True)
    self.assertEqual(obj.i, "abc")

  def test_instantiate_indefinite_order(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Make sure the class is instantiated.
    * Specify all required parameters.
    * The order of the required parameters is the reverse of the definition.
    """
    obj, unused = Generator()._instantiate(ClassForTest, c = "c", b = "b", a = "a")
    self.assertEqual(obj.a, "a")
    self.assertEqual(obj.b, "b")
    self.assertEqual(obj.c, "c")
    self.assertEqual(obj.d, 1)
    self.assertEqual(obj.e, 2)
    self.assertEqual(obj.f, 3)
    self.assertEqual(obj.g, None)
    self.assertEqual(obj.h, False)
    self.assertEqual(obj.i, None)

  def test_instantiate_classparam(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Make sure that parameters starting with `_` and `::` are not treated as class values.
    * Specify all required parameters.
    """
    obj, unused = Generator()._instantiate(ClassForTest, **{
      "a": "a",
      "b": "b",
      "c": "c",
      "/d": "d",
      "::id": "e",
    })
    self.assertEqual(obj.a, "a")
    self.assertEqual(obj.b, "b")
    self.assertEqual(obj.c, "c")
    self.assertEqual(obj.d, 1)
    self.assertEqual(obj.e, 2)
    self.assertEqual(obj.f, 3)
    self.assertEqual(obj.g, None)
    self.assertEqual(obj.h, False)
    self.assertEqual(obj.i, None)
    self.assertFalse(hasattr(obj, "_d"))
    self.assertFalse(hasattr(obj, "::id"))

  def test_instantiate_no_require_params(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Confirm that TypeError occurs.
    * Mandatory parameter not specified.
    """
    with self.assertRaises(TypeError):
      Generator._instantiate(ClassForTest)

  def test_instantiate_unknown_param(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Confirm that AttributeError occurs.
    * Specify all required parameters.
    * Unknown parameter is specified.
    """
    with self.assertRaises(AttributeError):
      Generator()._instantiate(ClassForTest, a = "a", b = "b", c = "c", z = 0)

  def test_instantiate_set_idtags(self):
    """
    When you call `Generator#_instantiate()` under the following conditions,
    Make sure the class is instantiated.
    * Specify all required parameters.
    * ID and tag elements are defined.
    """
    unused, tag = Generator()._instantiate(ClassForTest, **{
      "a": "a",
      "b": "b",
      "c": "c",
      "::id": "testid",
      "::tag": "testtag"})
    self.assertEqual(tag.id, "testid")
    self.assertEqual(tag.tag, "testtag")

  #endregion

if __name__ == "__main__":
  unittest.main()


class EventReciever(object):
  def __init__(self, object, tag, callback):
    self.object = object
    self.tag = tag
    self.callback = callback

  def __call__(self, *args, **kw):
    name = self.tag if args == () else args[0]
    self.callback(self.object, name)


import unittest
import tkinter
import yaml

from tksugar.generator import Generator
from tksugar.widgets import GeneratorSupport

class TestChildObject(object):
  def __init__(self, a):
    self.a = a

class TestOwner(GeneratorSupport):
  def __init__(self):
    self.children = []

  def append_child(self, child, **params):
    self.children.append({
      "obj": child,
      "param": params
    })

class Test_GeneratorSupport(unittest.TestCase):
  """
  Tests the `GeneratorSupport` Class

  Note: We're using a test class here to test the behavior, but we're using the
  Because of the import, the class name of the object generated by the Generator and the class name of the test class that can be referenced internally are different.
  Therefore, this test class cannot match the value of `type(actual) is expected`.
  """

  def test_append_child(self):
    """
    When the `Generator#generate()` method is called with the following conditions
    Make sure that the child object calls the `GeneratorSupport#append_child()` of the owner object.
    * The owner object implements `GeneratorSupport#append_child()`.
    """
    obj = Generator(file="tests/definition/generator_test/support_haschild.yml", modules=["tests.test_generatorsupport"]).generate()
    self.assertEqual(obj.__class__.__name__, "TestOwner")
    self.assertEqual(obj.children[0]["obj"].__class__.__name__, "TestChildObject")
    self.assertEqual(obj.children[0]["obj"].a, 1)
    self.assertEqual(obj.children[0]["param"]["a"], 1)
    self.assertEqual(obj.children[0]["param"]["b"], "a")

if __name__ == "__main__":
  unittest.main()